/*****************************************************************************
 * loadmap.cpp - map class, the functions to load the map goes here
 *
 ****************************************************************************/
#include "SDL.h"

#include "freecnc.h"
#include "map.h"
#include "vfs.h"
#include "inifile.h"
#include "shpimage.h"
#include "endian.h"

#include "game.h"

#include "memtrack.h"

#include "compression.h"


#include <string>

using namespace std;



/** Loads the maps ini file containing info on dimentions, units, trees
 * and so on. 
 * @param the mixfiles.
 */

void CnCMap::loadIni() {
    INIFile *inifile;
    char *ininame = new char[strlen(mapname)+5];

    /* The name of the ini is mapname.ini */
    strcpy(ininame, mapname);
    strcat(ininame, ".INI");

    /* Find and seek the ini file in the mix */
    try {
        inifile = new INIFile(ininame);
    } catch (INInotFound *e) {
        delete e;
        fprintf(stderr, "Mission not in MIX file (or size is zero)\nPlease make sure you have the correct (NOD or GDI) general.mix in the right place\n" );
        /* throw an error */
        exit(1);
    }

    delete[] ininame;

    if( inifile->readInt("basic", "newiniformat", 0) != 0 ) {
        printf("Red Alert maps not supported yet\n");
        //return;
    }

    simpleSections(inifile);


    uspool = new UnitAndStructurePool(this, evQueue);

    terraintypes = new Uint8[width*height];
    memset(terraintypes, 0, width*height);

    advancedSections(inifile);

    if (maptype == GAME_RA)
        unMapPack(inifile);

    delete inifile;

}



/** Function to load all vars in the simple sections of the inifile
 
 * @param a pointer to the inifile
 
 */

void CnCMap::simpleSections(INIFile *inifile) {

    /* The strings in the basic section */

    brief = inifile->readString("BASIC", "BRIEF");
    action = inifile->readString("BASIC", "ACTION");
    player = inifile->readString("BASIC", "PLAYER");
    theme = inifile->readString("BASIC", "THEME");
    winmov = inifile->readString("BASIC", "WIN");
    losemov = inifile->readString("BASIC", "LOSE");

    /* The ints (and one string) in the map section */

    height = inifile->readInt("MAP", "HEIGHT");
    width = inifile->readInt("MAP", "WIDTH");
    x = inifile->readInt("MAP", "X");
    y = inifile->readInt("MAP", "Y");
    theater = inifile->readString("MAP", "THEATER");

    if( brief == NULL || action == NULL || player == NULL || theater == NULL ||
            theme == NULL || winmov == NULL || losemov == NULL ) {
        fprintf(stderr, "Error loading map\n");
        exit(1); /* throw error */
    }

    if( height == INIERROR || width == INIERROR || x == INIERROR || y == INIERROR ) {
        fprintf(stderr, "Error loading map\n");
        exit(1); /* throw exeption */
    }
}



/** Function to load all the advanced sections in the inifile.
 
 * @param a pointer to the inifile
 
 */

void CnCMap::advancedSections(INIFile *inifile) {
    int i;
    char *line;
    char shpname[128];
    char trigger[128];
    char action[128];
    char type[128];
    char owner[128];
    Uint32 facing, health, subpos;
    Uint32 linenum, smudgenum, tmpval;
    Uint16 tx, ty, xsize, ysize;
    map<string, unsigned int> imagelist;
    map<string, unsigned int>::iterator imgpos;
    SHPImage *image;
    TerrainEntry tmpterrain;

    Uint16 xwalk, ywalk, ttype;

    INIFile *arts = new INIFile("art.ini");

    /* load the smudge marks and the tiberium to the imagepool */
    if( maptype == GAME_TD ) {
        strncpy(shpname,"TI1.",128);
        strncat(shpname,theater,3);
        try { // TI1.SNO does not exist
            image = new SHPImage(shpname, Game::mapscaleq);
            imagepool.push_back(image);
        } catch (ImageNotFound* e) {
            delete e;
            printf("Could not load %s\n",shpname);
        }
    } else if( maptype == GAME_RA ) {

        strcpy(shpname, "GOLD01."); // Red Alert Ore
        strncat(shpname, theater, 3);
        try {
            image = new SHPImage(shpname, Game::mapscaleq);
            imagepool.push_back(image);
        } catch (ImageNotFound* e) {
            delete e;
            printf("Could not load %s\n",shpname);
        }
    } else {
        printf("Unsuported maptype\n");
    }

    for( i = 1; i <= 6; i++ ) {
        sprintf(shpname, "SC%d.", i);
        strncat(shpname, theater, 3);
        image = new SHPImage(shpname, Game::mapscaleq);
        imagepool.push_back(image);
    }
    for( i = 1; i <= 6; i++ ) {
        sprintf(shpname, "CR%d.", i);
        strncat(shpname, theater, 3);
        image = new SHPImage(shpname, Game::mapscaleq);
        imagepool.push_back(image);
    }

    overlaymatrix = new Uint16[width*height];
    memset(overlaymatrix, 0, sizeof(Uint16)*width*height);

    inifile->seekSection("TERRAIN");
    line = inifile->getLine();
    while( line != NULL ) {
        /* , is the char which separate terraintype from action. */
        if( sscanf(line, "%d=%[^,],%s", &linenum, shpname, trigger) == 3 ) {
            /* Set the next entry in the terrain vector to the correct values.
             * the map-array and shp files vill be set later */
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }

            if( shpname[0] == 't' || shpname[0] == 'T' )
                ttype = TTYPE_TREE;
            else if( shpname[0] == 'r' || shpname[0] == 'R' )
                ttype = TTYPE_ROCK;
            else
                ttype = TTYPE_OTHER_NONPASS;

            /* calculate the new pos based on size and blocked */
            xsize = arts->readInt(shpname, "xsize",1);
            ysize = arts->readInt(shpname, "ysize",1);

            for( ywalk = 0; ywalk < ysize && ywalk + ty < height+y; ywalk++ ) {
                for( xwalk = 0; xwalk < xsize && xwalk + tx < width + x; xwalk++ ) {
                    sprintf(type, "notblocked%d", ywalk*xsize+xwalk);
                    if( arts->readInt(shpname, type) == INIERROR ) {
                        terraintypes[(ywalk+ty-y)*width+xwalk+tx-x] = ttype;
                    }
                }
            }

            linenum = xsize*ysize;
            do {
                linenum--;
                sprintf(type, "notblocked%d", linenum);
            } while(arts->readInt(shpname, type) == INIERROR);


            tmpterrain.xoffset = -(linenum%ysize)*24;
            tmpterrain.yoffset = -(linenum/ysize)*24;

            tx += linenum%ysize;
            if( tx >= width+x ) {
                tmpterrain.xoffset += 1+tx-(width+x);
                tx = width+x-1;
            }
            ty += linenum/ysize;
            if( ty >= height+y ) {
                tmpterrain.yoffset += 1+ty-(height+y);
                ty = height+y-1;
            }



            linenum = (ty-y)*width + tx - x;
            strcat(shpname, ".");
            strncat(shpname, theater, 3);


            /* search the map for the image */
            imgpos = imagelist.find(shpname);

            /* set up the overlay matrix and load some shp's */
            if( imgpos != imagelist.end() ) {

                /* this tile already has a number */
                overlaymatrix[linenum] |= HAS_TERRAIN;
                tmpterrain.shpnum = imgpos->second << 16;
                terrains[linenum] = tmpterrain;
            } else {
                /* a new tile */
                imagelist[shpname] = imagepool.size();
                overlaymatrix[linenum] |= HAS_TERRAIN;
                tmpterrain.shpnum = imagepool.size()<<16;
                terrains[linenum] = tmpterrain;
                image = new SHPImage(shpname, Game::mapscaleq);
                imagepool.push_back(image);
            }
        }
        delete[] line;
        line = inifile->getLine();
    }
    delete[] line;

    inifile->seekSection("OVERLAY");
    line = inifile->getLine();
    while( line != NULL ) {
        if( sscanf(line, "%d=%s", &linenum, shpname) == 2 ) {
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }
            linenum = (ty-y)*width + tx - x;

            /* Set the next entry in the overlay vector to the correct values.
             * the map-array and shp files vill be set later */
            if( !strcmp(shpname, "BRIK") || !strcmp(shpname, "SBAG") ||
                    !strcmp(shpname, "BARB") || !strcmp(shpname, "WOOD") ||
                    !strcmp(shpname, "CYCL") ) {

                /* Make the walls structures instead */
                uspool->createStructure(shpname, linenum, ppool.getPlayerNum("neutral", inifile));
            } else if( !strncmp(shpname, "TI", 2) ) {
                /* tiberium */
                sscanf(shpname,"TI%i",&i);
                overlaymatrix[linenum] |= (i-1);
            } else {
                strcat(shpname, ".SHP");

                /* search the map for the image */
                imgpos = imagelist.find(shpname);

                /* set up the overlay matrix and load some shp's */
                if( imgpos != imagelist.end() ) {

                    /* this tile already has a number */
                    overlaymatrix[linenum] |= HAS_OVERLAY;
                    overlays[linenum] = imgpos->second;
                } else {

                    /* a new tile */
                    imagelist[shpname] = imagepool.size();
                    overlaymatrix[linenum] |= HAS_OVERLAY;
                    overlays[linenum] = imagepool.size();
                    try {
                        image = new SHPImage(shpname, Game::mapscaleq);
                    } catch(ImageNotFound *e2) {
                        delete e2;
                        shpname[strlen(shpname)-3] = '\0';
                        strncat(shpname, theater, 3);
                        image = new SHPImage(shpname, Game::mapscaleq);
                    }
                    imagepool.push_back(image);
                }

                if( shpname[0] == 't' || shpname[0] == 'T' )
                    terraintypes[linenum] = TTYPE_TREE;
                else if( shpname[0] == 'r' || shpname[0] == 'R' )
                    terraintypes[linenum] = TTYPE_ROCK;
                else
                    terraintypes[linenum] = TTYPE_OTHER_NONPASS;
            }

        }
        delete[] line;
        line = inifile->getLine();
    }
    delete[] line;

    inifile->seekSection("SMUDGE");
    line = inifile->getLine();
    while( line != NULL ) {

        /* , is the char which separate terraintype from action. */
        if( sscanf(line, "%d=SC%d", &linenum, &smudgenum) == 2 ) {
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }
            linenum = (ty-y)*width + tx - x;
            overlaymatrix[linenum] |= (smudgenum<<4);
        } else if( sscanf(line, "%d=CR%d", &linenum, &smudgenum) == 2 ) {
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }

            linenum = (ty-y)*width + tx - x;
            overlaymatrix[linenum] |= ((smudgenum+6)<<4);
        }
        delete[] line;
        line = inifile->getLine();
    }
    delete[] line;

    inifile->seekSection("STRUCTURES");
    line = inifile->getLine();
    while( line != NULL ) {
        /* , is the char which separate terraintype from action. */
        if( sscanf(line, "%d=%[^,],%[^,],%d,%d,%d,%s", &tmpval, owner, type,
                   &health, &linenum, &facing, trigger ) == 7  ) {
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }
            linenum = (ty-y)*width + tx - x;
            uspool->createStructure(type, linenum, ppool.getPlayerNum(owner, inifile));
        }
        delete[] line;
        line = inifile->getLine();
    }
    delete[] line;

    inifile->seekSection("UNITS");
    line = inifile->getLine();
    while( line != NULL ) {
        /* , is the char which separate terraintype from action. */
        if( sscanf(line, "%d=%[^,],%[^,],%d,%d,%d,%[^,],%s", &tmpval, owner, type,
                   &health, &linenum, &facing, action, trigger ) == 8  ) {
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }
            linenum = (ty-y)*width + tx - x;
            uspool->createUnit(type, linenum, 5, ppool.getPlayerNum(owner, inifile));
        }
        delete[] line;
        line = inifile->getLine();
    }
    delete[] line;

    /*infantry*/
    inifile->seekSection("INFANTRY");
    line = inifile->getLine();
    while( line != NULL ) {
        /* , is the char which separate terraintype from action. */
        if( sscanf(line, "%d=%[^,],%[^,],%d,%d,%d,%[^,],%d,%s", &tmpval, owner, type,
                   &health, &linenum, &subpos, action, &facing, trigger ) == 9  ) {
            tx = linenum%64;
            ty = linenum/64;
            if( tx < x || ty < y || tx > x+width || ty > height+y ) {
                delete[] line;
                line = inifile->getLine();
                continue;
            }
            linenum = (ty-y)*width + tx - x;
            uspool->createUnit(type, linenum, subpos, ppool.getPlayerNum(owner, inifile));
        }
        delete[] line;
        line = inifile->getLine();
    }
    delete[] line;
    delete arts;
}



/** Load the bin part of the map. This contains the tiles.
 */

struct tiledata {
    Uint32 image;
    Uint8 type;
};

void CnCMap::loadBin() {
    Uint32 index = 0;
    //    Uint8 templ, tile;
    int xtile, ytile;
    VFile *binfile;
    char *binname = new char[strlen(mapname)+5];

    TileList *mapdata;

    mapdata = new TileList[width*height];

    /* Calculate name of bin file ( mapname.bin ). */
    strcpy(binname, mapname);
    strcat(binname, ".BIN");

    /* get the offset and size of the binfile along with a pointer to it */
    //binfile = mixes->getOffsetAndSize(binname, &offset, &size);
    binfile = VFS_Open(binname);
    delete[] binname;

    if(binfile == NULL) {
        fprintf(stderr, "Unable to locate BIN file!\n");
        exit(1); /* should throw an error */
    }

    /* Seek the beginning of the map.
     * It's at begining of bin + maxwidth * empty y cells + empty x cells
     * times 2 sinse each entry is 2 bytes
     */
    binfile->seekSet( (64*y + x) * 2 );

    for( ytile = 0; ytile < height; ytile++ ) {
        for( xtile = 0; xtile < width; xtile++ ) {
            /* Read template and tile */
            mapdata[index].templateNum = 0;
            binfile->readByte((Uint8 *)&(mapdata[index].templateNum), 1);
            binfile->readByte(&(mapdata[index].tileNum), 1);

            //	      fread(mapdata+index+1, 1, 1, binfile);
            index++;
        }
        /* Skip til the end of the line and the onwards to the
         * beginning of usefull data on the next line 
         */
        binfile->seekCur( 2*(64-width) );
    }
    VFS_Close(binfile);
    parseBin(mapdata);
    delete[] mapdata;
}

void CnCMap::unMapPack( INIFile* inifile) {
    Uint32 tmpval, curpos;
    int xtile, ytile;
    TileList *bindata;
    char* line;
    //char *line2 = new char[49152]; // 48k
    Uint8 *mapdata1 = new Uint8[49152]; // 48k
    Uint8 *mapdata2 = new Uint8[49152];
    //memset(mapdata,0,49152);

    // read packed data into array
    inifile->seekSection("MAPPACK");
    ((char*)mapdata1)[0] = '\0';
    while ( 1 ) {
        line = inifile->getLine();
        if( line != NULL ) {
            if( sscanf(line,"%d=%s",&tmpval,line) == 2 ) {
                strcat(((char*)mapdata1), line);
            }
            delete[] line;
        } else {
            break;
        }
    }
    Compression::dec_base64(mapdata1,mapdata2,strlen(((char*)mapdata1)));

    /* decode the format80 coded data (6 chunks) */
    curpos = 0;
    for( tmpval = 0; tmpval < 6; tmpval++ ) {
        //printf("first vals in data is %x %x %x %x\n", mapdata2[curpos],
        //mapdata2[curpos+1], mapdata2[curpos+2], mapdata2[curpos+3]);
        if( Compression::decode80((Uint8 *)mapdata2+4+curpos, mapdata1+8192*tmpval) != 8192 ) {
            printf("A format80 chunk in the \"MapPack\" was of wrong size\n");
        }
        curpos = curpos + 4 + mapdata2[curpos] + (mapdata2[curpos+1]<<8) +
                 (mapdata2[curpos+2]<<16);
    }
    delete[] mapdata2;
    //printf("size of first chunk was %d\n", Compression::decode80((Uint8 *)mapdata+4, mapdata2));
    /*
         inifile->seekSection("OVERLAYPACK");
         line = inifile->getLine();
         sscanf(line,"%d=%s",&tmpval,line);
         while ( line != NULL ) {
    	  printf("%s\n",line);
    	  delete[] line;
    	  line = inifile->getLine();
    	  sscanf(line,"%d=%s",&tmpval,line);
         }
    */
    /* 128*128 16-bit template number followed by 128*128 8-bit tile numbers */
    bindata = new TileList[width*height];
    tmpval = y*128+x;
    curpos = 0;
    for( ytile = 0; ytile < height; ytile++ ) {
        for( xtile = 0; xtile < width; xtile++ ) {
            /* Read template and tile */
            bindata[curpos].templateNum = ((Uint16 *)mapdata1)[tmpval];
            bindata[curpos].tileNum = mapdata1[tmpval+128*128*2];
            curpos++;
            tmpval++;
            /*		printf("tile %d, %d\n", bindata[curpos-1].templateNum,
            					bindata[curpos-1].tileNum); */
        }
        /* Skip til the end of the line and the onwards to the
         * beginning of usefull data on the next line 
         */
        tmpval += (128-width);
    }
    delete[] mapdata1;
    parseBin(bindata);
    delete[] bindata;
}

void CnCMap::parseBin(TileList* bindata) {
    Uint32 index;

    Uint16 templ;
    Uint8 tile;
    int xtile, ytile;
    INIFile *templini;

    SDL_Surface *tileimg;
    SDL_Color palette[256];

    map<Uint32, struct tiledata> tilelist;
    map<Uint32, struct tiledata>::iterator imgpos;
    //     char tempc[sizeof(Uint8)];

    struct tiledata tiledata;
    Uint32 tiletype;
    tilematrix = new Uint16[width*height];

    loadPal(palette);
    SHPBase::setPalette(palette);
    SHPBase::calculatePalettes();

    /* Load the templates.ini */
    try {
        if( maptype == GAME_RA ) {
            templini = new INIFile("ratemplates.ini");
        } else {
            templini = new INIFile("templates.ini");
        }
    } catch( INInotFound *e ) {
        delete e;
        exit(1);
    }
    index = 0;
    for( ytile = 0; ytile < height; ytile++ ) {
        for( xtile = 0; xtile < width; xtile++ ) {
            /* Read template and tile */
            templ = bindata[index].templateNum;
            tile  = bindata[index].tileNum;
            index++;
            /* Template 0xff is an empty tile */
            if(templ == ((maptype == GAME_RA)?0xffff:0xff)) {
                templ = 0;
                tile = 0;
            }

            /* Code sugested by Olaf van der Spek to cause all tiles in template
             * 0 and 2 to be used */
            if( templ == 0)
                tile = xtile&3 | (ytile&3 << 2);
            else if( templ == 2 )
                tile = xtile&1 | (ytile&1 << 1);


            imgpos = tilelist.find(templ<<8 | tile);

            /* set up the tile matrix and load some tiles */
            if( imgpos != tilelist.end() ) {
                /* this tile already has a number */
                tilematrix[width*ytile+xtile] = imgpos->second.image;
                if( terraintypes[width*ytile+xtile] == 0 )
                    terraintypes[width*ytile+xtile] = imgpos->second.type;
            } else {
                /* a new tile */
                tileimg = loadTile(templini, templ, tile, tiletype);

                tiledata.image = tileimages.size();
                tiledata.type = tiletype;
                tilelist[templ<<8 | tile] = tiledata;
                tilematrix[width*ytile+xtile] = tileimages.size();
                if( terraintypes[width*ytile+xtile] == 0 )
                    terraintypes[width*ytile+xtile] = tiletype;

                if( tileimg == NULL ) {
                    fprintf(stderr, "Error loading tiles\n");
                    exit(1); /* throw an exception */
                }

                tileimages.push_back(tileimg);
            }

        }
    }
}

/** Load a palette
 * @param the mixfiles.
 * @param a structure to store the palette in.
  */

void CnCMap::loadPal( SDL_Color *palette ) {
    VFile *mixfile;
    int i;
    char *palname = new char[strlen(theater)+5];

    /* Calculate the name of the palette ( theater.pal ) */
    strncpy(palname, theater, 8);
    palname[8] = '\0';
    strcat(palname, ".PAL");

    /* Seek the palette file in the mix */
    mixfile = VFS_Open(palname);

    if(mixfile == NULL) {
        fprintf(stderr, "Unable to locate palette! This shouldn't happen unles you do something stupid\n"
                "name: %s\n",palname);
        //	abort();
        return;
    }

    delete[] palname;


    /* Load the palette */
    for (i = 0; i < 256; i++) {
        mixfile->readByte(&palette[i].r, 1);
        mixfile->readByte(&palette[i].g, 1);
        mixfile->readByte(&palette[i].b, 1);
        palette[i].r <<= 2;
        palette[i].g <<= 2;
        palette[i].b <<= 2;
    }
    VFS_Close(mixfile);
    /* I have no idea how shadows work, so I set the "shadow color" to a more
     * apropiate color than brightgreen. And make it translucent */
    palette[4].r = 5;
    palette[4].g = 5;
    palette[4].b = 5;
}



/** load a tile from the mixfile.
 * @param the mixfiles.
 * @param the template inifile.
 * @param the template number.
 * @param the tilenumber.
 * @returns a SDL_Surface containing the tile.
 */
SDL_Surface *CnCMap::loadTile( INIFile *templini,
                               Uint16 templ, Uint8 tile, Uint32 &tiletype ) {

    TemplateImage *theaterfile;

    SDL_Surface *retimage;


    char tilefilename[13];
    char tilenum[11];
    char *temname;


    /* The name of the file containing the template is something from
     * templates.ini . the three first
     * chars in the name of the theater eg. .DES .TEM .WIN */

    sprintf(tilefilename, "TEM%d", templ);

    sprintf(tilenum, "tiletype%d", tile);
    tiletype = templini->readInt(tilefilename, tilenum);
    if( tiletype == INIERROR )
        tiletype = 0;
    //   printf("Type is %d\n", tiletype);

    temname = templini->readString(tilefilename, "Name");

    if( temname == NULL ) {
        fprintf(stderr, "Error in templates.ini! (can't find \"%s\")\n", tilefilename);
        strcpy(tilefilename, "CLEAR1");
    } else {
        strcpy(tilefilename, temname);
        delete[] temname;
    }

    strcat( tilefilename, "." );
    strncat( tilefilename, theater, 3 );


    /* Find and seek the template file in the mix */

    try {
        if( maptype == GAME_RA ) {
            theaterfile = new TemplateImage(tilefilename, Game::mapscaleq, 1);
        } else {
            theaterfile = new TemplateImage(tilefilename, Game::mapscaleq);
        }
    } catch(ImageNotFound *e) {
        delete e;
        fprintf(stderr, "Unable to locate template %d, %d (\"%s\") in mix! using tile 0, 0 instead\n", templ, tile, tilefilename);
        if( templ == 0 && tile == 0 )
            return NULL;
        return loadTile( templini, 0, 0, tiletype );
    }


    retimage = theaterfile->getImage(tile);
    if( retimage == NULL ) {
        fprintf(stderr, "Illegal template %d, %d (\"%s\")! using tile 0, 0 instead\n", templ, tile, tilefilename);
        if( templ == 0 && tile == 0 )
            return NULL;
        return loadTile( templini, 0, 0, tiletype );
    }
    delete theaterfile;
    return retimage;

}

