/*****************************************************************************
 * map.cpp - map class, the functions to handle a already loaded map goes here
 *
 ****************************************************************************/

#include "SDL.h"

#include "freecnc.h"
#include "map.h"

#include "path.h"
#include "memtrack.h"

using namespace std;


/** Get the imagepool, called by the imagecache.
 * @returns the imagepool.
 */

vector<SHPImage *> *CnCMap::getImagePool() {
    return &imagepool;
}


Uint32 CnCMap::getOverlay(Uint32 pos) {
    Uint32 overlay;

    if( overlaymatrix[pos] & HAS_OVERLAY ) {
        overlay = overlays[pos]<<16;
    } else
        overlay = 0;
    return overlay;
}

Uint32 CnCMap::getTerrain(Uint32 pos, Sint16 &xoff, Sint16 &yoff) {
    Uint32 terrain;

    if( overlaymatrix[pos] & HAS_TERRAIN ) {
        terrain = terrains[pos].shpnum;
        xoff = terrains[pos].xoffset;
        yoff = terrains[pos].yoffset;
    } else
        terrain = 0;
    return terrain;
}



/** Constructor that loads a map cmoplete with terrain, overlays, units
 * and structures.
 * @param the mixfiles.
 * @param the eventqueue.
 * @param the name of the map.
 */
CnCMap::CnCMap( ActionEventQueue *evQueue, char *mapname, Uint8 maptype ) {
    /* Copy the name of the map so we can store it */

    strncpy( this->mapname, mapname, 9 );
    this->evQueue = evQueue;
    this->maptype = maptype;

    /* Load the ini part of the map */
    fprintf(stderr,"\t\tReading %s.INI\n",mapname);
    loadIni();

    /* Load the bin part of the map ( the tiles ) */
    fprintf(stderr,"\t\tLoading %s.BIN\n",mapname);
    if (maptype == GAME_TD)
        loadBin();
    /* start at top right corner of map. */
    /** FIXME: get start map position from MAPNAME.ini */
    ypos = 0;
    xpos = 0;

    scrollvec.x = 0;
    scrollvec.y = 0;
    scrollvec.t = 0;

    //   Path::setMapSize(width, height);
    ppool.setHPlayer(player);
    minimap = NULL;
}



/** Destructor, free up some memory */
CnCMap::~CnCMap() {
    unsigned int i;

    delete[] brief;
    delete[] action;
    delete[] player;
    delete[] theater;
    delete[] theme;
    delete[] winmov;
    delete[] losemov;

    delete[] tilematrix;
    delete[] terraintypes;

    for( i = 0; i < tileimages.size(); i++ )
        SDL_FreeSurface(tileimages[i]);

    delete[] overlaymatrix;

    for( i = 0; i < imagepool.size(); i++ )
        delete imagepool[i];

    delete uspool;
}



/** sets the scroll to the specified direction.
 * @param direction to scroll in.
 */

void CnCMap::scroll(Uint8 direction) {
    if (direction & SCROLL_UP) {
        if (scrollvec.y >= 0)
            scrollvec.y = -SCROLL_DELTA;
        else if (scrollvec.y > -MAXSCROLL)
            scrollvec.y -= SCROLL_DELTA;
    }
    if (direction & SCROLL_DOWN) {
        if (scrollvec.y <= 0)
            scrollvec.y = SCROLL_DELTA;
        else if (scrollvec.y < MAXSCROLL)
            scrollvec.y += SCROLL_DELTA;
    }
    if (direction & SCROLL_LEFT) {
        if (scrollvec.x >= 0)
            scrollvec.x =  -SCROLL_DELTA;
        else if (scrollvec.x > -MAXSCROLL)
            scrollvec.x -= SCROLL_DELTA;
    }
    if (direction & SCROLL_RIGHT) {
        if (scrollvec.x <= 0)
            scrollvec.x = SCROLL_DELTA;
        else if (scrollvec.x < MAXSCROLL)
            scrollvec.x += SCROLL_DELTA;
    }
    scrollvec.t = 0;
    toscroll = true;
}



/** scrolls according to the scroll vector.
 */
void CnCMap::doscroll() {
    if ((xpos <= SCROLL_DELTA) && (scrollvec.x < 0)) {
        scrollvec.x = 0;
        xpos = 0;
        scrollvec.t = 0;
    }
    if ((ypos <= SCROLL_DELTA) && (scrollvec.y < 0)) {
        scrollvec.y = 0;
        ypos = 0;
        scrollvec.t = 0;
    }
    if ((xpos >= x)&&(scrollvec.x > 0)) {
        xpos = x;
        scrollvec.x = 0;
    }
    if ((ypos >= y)&&(scrollvec.y > 0)) {
        ypos = y;
        scrollvec.y = 0;
    }
    if ((scrollvec.x == 0) && (scrollvec.y == 0)) {
        /* normalise [xy]pos anyway */
        if (xpos > x)
            xpos = x;
        if (ypos > y)
            ypos = y;
        toscroll = false;
        return;
    }
    xpos += scrollvec.x;
    ypos += scrollvec.y;
    ++scrollvec.t;
    /* scrolling continues at current rate for MAXTSCROLL
     * passes then decays quickly */
    if (scrollvec.t >= MAXTSCROLL) {
        scrollvec.x /=2;
        scrollvec.y /=2;
    }
}



/** sets the maximum value the scroll can take on.
 * @param the maximum x scroll.
 * @param the maximum y scroll.
 */

void CnCMap::setMaxScroll( Uint32 x, Uint32 y ) {
    this->x = 0;
    this->y = 0;
    if( width > x )
        this->x = width - x;
    if( height > y )
        this->y = height - y;

    if( xpos > this->x )
        xpos = this->x;

    if( ypos > this->y )
        ypos = this->y;

}

Uint16 CnCMap::getCost(Uint16 pos) {
    return getCost(pos,NULL);
}

Uint16 CnCMap::getCost(Uint16 pos,Unit* excpUn) {
    Uint16 cost;

    switch(terraintypes[pos]) {
    case TTYPE_ROCK:
    case TTYPE_TREE:
    case TTYPE_WATER:
    case TTYPE_WATER_BLOCKED:
    case TTYPE_OTHER_NONPASS:
        cost = 0xffff;
        break;
    default:
        cost = 1 + uspool->getTileCost(pos,excpUn);
        break;
    }
    return cost;
}

SDL_Surface* CnCMap::getMiniMap() {
    SDL_Surface *cminitile;
    if(minimap == NULL) {
        minimap = SDL_CreateRGBSurface (SDL_SWSURFACE, width*6,height*6, 16, 0xff, 0xff, 0xff, 0);
        SDL_Rect *pos = new SDL_Rect();
        ImageProc *ip = new ImageProc();
        pos->w = 6;
        pos->h = 6;
        pos->x = 0;
        pos->y = 0;
        SDL_Surface *maptileOne = getMapTile(0);
        minimap->format->Rmask = maptileOne->format->Rmask;
        minimap->format->Gmask = maptileOne->format->Gmask;
        minimap->format->Bmask = maptileOne->format->Bmask;
        minimap->format->Amask = maptileOne->format->Amask;
        if( maptileOne->format->palette != NULL ) {
            SDL_SetColors(minimap, maptileOne->format->palette->colors, 0, maptileOne->format->palette->ncolors);
        }
        int lineCounter = 0;
        for(Uint32 i = 0;  i < (Uint32) width*height; i++, pos->x += 6, lineCounter++) {
            if(lineCounter == width) {
                pos->y += 6;
                pos->x =0;
                lineCounter = 0;
            }
            cminitile = ip->minimapScale(getMapTile(i));
            SDL_BlitSurface(cminitile, NULL, minimap, pos);
            SDL_FreeSurface(cminitile);
        }
    }
    return minimap;

}
/*
Uint8 CnCMap::getSide(void) {
    if (strcasecmp("goodguy",player)) {
        return 1;
    } else
        return 0;
}
*/
void CnCMap::printPlayersUnitsStructures() { // only for development purposes
    // for each player
    for(Uint8 i = 0; i < ppool.getNumPlayers(); i++) {
        Player *aPlayer = ppool.getPlayer(i);
        cerr << "Player " << (int)i << " Name " << aPlayer->getName() << " Side " << aPlayer->getSide() << " PaletteNum " << (int)aPlayer->getPaletteNum() << endl;
        cerr << "  Power " << aPlayer->getPower() << " of PowerUsed " << aPlayer->getPowerUsed() << endl;
        // for all its units
        vector<Unit *> aUnits = aPlayer->getUnits();
        for (vector<Unit *>::iterator aUnitsI = aUnits.begin();
                aUnitsI != aUnits.end();
                aUnitsI++) {
            UnitType *aUnitType = (UnitType *)(*aUnitsI)->getType();
            cerr << "  Unit at Pos " << (*aUnitsI)->getPos() << " Subpos " << (*aUnitsI)->getSubpos() << " Health " << (*aUnitsI)->getHealth() << " of MaxHealth " << (int)aUnitType->getMaxHealth() << " Cost " << (int)aUnitType->getCost() << endl;
            cerr << "    Unitnum " << (*aUnitsI)->getUnitnum() << " isInfantry " << (int)aUnitType->isInfantry() << endl;
            cerr << "    Speed " << (int)aUnitType->getSpeed() << " Turnspeed " << (int)aUnitType->getTurnspeed() << " Sight " << (int)aUnitType->getSight() << " ROT " << (int)aUnitType->getROT() << " MoveMod " << (int)aUnitType->getMoveMod() << " TurnMod " << (int)aUnitType->getTurnMod() << endl;
        }
        // for all its structures
        vector<Structure *> aStructures = aPlayer->getStructures();
        for (vector<Structure *>::iterator aStructuresI = aStructures.begin();
                aStructuresI != aStructures.end();
                aStructuresI++) {
            StructureType *aStructureType = (StructureType *)(*aStructuresI)->getType();
            cerr << "  Structure at Pos " << (*aStructuresI)->getPos() << " Subpos " << (*aStructuresI)->getSubpos() << " Health " << (*aStructuresI)->getHealth() << " of MaxHealth " << (int)aStructureType->getMaxHealth() << " Cost " << (int)aStructureType->getCost() << endl;
            cerr << "    isWall " << (int)aStructureType->isWall() << " Turnspeed " << (int)aStructureType->getTurnspeed() << " Sight " << (int)aStructureType->getSight() << endl;
        }
    }
}
