#include <math.h>

#include "projectileanim.h"
#include "unitorstructure.h"
#include "weaponspool.h"
#include "unitandstructurepool.h"
#include "memtrack.h"

ExplosionAnim::ExplosionAnim(Uint32 p, UnitAndStructurePool* uspool, Uint16 pos,
                             Uint32 startimage, Uint8 animsteps, Sint8 xoff=0, Sint8 yoff=0) : ActionEvent(p) {
    this->uspool = uspool;
    l2o = new L2Overlay();
    l2o->imagenum = startimage;
    l2o->xoffset = xoff;
    l2o->yoffset = yoff;
    l2o->cellpos = pos;
    l2entry = uspool->addL2overlay(pos, l2o);
    this->animsteps = animsteps;
    uspool->scheduleUSEvent(this);
    this->pos = pos;
}

ExplosionAnim::~ExplosionAnim() {
    uspool->removeL2overlay(l2entry);
    delete l2o;
}

void ExplosionAnim::run() {
    animsteps--;
    if( animsteps == 0 ) {
        delete this;
        return;
    }
    l2o->imagenum++;
    uspool->scheduleUSEvent(this);
}

ProjectileAnim::ProjectileAnim(Uint32 p, Weapon *weap, UnitOrStructure* owner, Uint16 dest, Uint8 subdest, UnitAndStructurePool *uspool) : ActionEvent(p) {
    double pixelspertick;
    double totlen;
    float alpha;

    this->weap = weap;
    this->owner = owner;
    this->dest = dest;
    this->subdest = subdest;
    this->uspool = uspool;
    owner->referTo();

    heatseek = weap->isHeatseek();
    inaccurate = weap->isInaccurate();
    facing = 0;

    if (heatseek) {
        target = uspool->getUnitOrStructureAt(dest,subdest);
        if (target == NULL) {
            //fprintf(stderr,"Could not get target from uspool reverting to non heat-seeking behaviour\n");
            heatseek = false;
        } else {
            target->referTo();
        }
    } else {
        target = NULL;
    }
    // xdiff, ydiff
    // xmod, ymod
    if( weap->getSpeed() < 100 ) {
        //printf("speed is < 100\n");
        pixelspertick = (double)weap->getSpeed()/4.0;
        xdiff = owner->getPos()%uspool->getMap()->getWidth() - dest%uspool->getMap()->getWidth();
        ydiff = owner->getPos()/uspool->getMap()->getWidth() - dest/uspool->getMap()->getWidth();
        xdiff *= 24;
        ydiff *= 24;
        totlen = sqrt(xdiff*xdiff+ydiff*ydiff);
        xmod = -(xdiff*pixelspertick)/totlen;
        ymod = -(ydiff*pixelspertick)/totlen;
        rxoffs = owner->getXoffset();
        ryoffs = owner->getYoffset();
        //printf("%d, %d == %lf, %lf\n", owner->getXoffset(), owner->getYoffset(), rxoffs, ryoffs);

        if (weap->getProjectile()->doesRotate()) {
            if( xdiff == 0 ) {
                if( ydiff < 0 ) {
                    alpha = -3.1415926536f/2.0f;
                } else {
                    alpha = 3.1415926536f/2.0f;
                }
            } else {
                alpha = atan((float)ydiff/(float)xdiff);
                if( xdiff < 0 ) {
                    alpha = 3.1415926536f+alpha;
                }
            }
            facing = (40-(Sint8)(alpha*16/3.1415926536f))&0x1f;
        } else {
            facing = 0;
        }

        l2o = new L2Overlay();
        l2o->cellpos = owner->getPos();
        l2o->imagenum = weap->getProjectile()->getImageNum()+facing;
        l2o->xoffset = owner->getXoffset();
        l2o->yoffset = owner->getYoffset();
        l2entry = uspool->addL2overlay(owner->getPos(), l2o);
        //setDelay(1);
        //printf("diff: %d, %d. mod: %d, %d - totlen: %d pix/tick: %d\n", xdiff, ydiff, xmod, ymod, totlen, pixelspertick);
    }
    else {
        // instant hit..
        xmod = 0;
        ymod = 0;
        l2o = NULL;
    }
}

ProjectileAnim::~ProjectileAnim() {
    if( l2o != NULL ) {
        uspool->removeL2overlay(l2entry);
        delete l2o;
    }
    if (heatseek) {
        if (target->isAlive()) {
            target->unrefer();
        }
    }
    owner->unrefer();
}

void ProjectileAnim::run() {
    Uint32 oldpos;
    Unit *utarget;
    Structure *starget;
    if (heatseek) {
        double pixelspertick;
        double totlen;
        float alpha;
        dest = target->getBPos();
        subdest = target->getSubpos();
        pixelspertick = (double)weap->getSpeed()/4.0;
        xdiff = (l2o->cellpos)%uspool->getMap()->getWidth() - dest%uspool->getMap()->getWidth();
        ydiff = (l2o->cellpos)/uspool->getMap()->getWidth() - dest/uspool->getMap()->getWidth();
        if ((xdiff == 0) && (ydiff == 0)) {
            xmod = 0;
            ymod = 0;
        } else {
            xdiff *= 24;
            ydiff *= 24;
            totlen = sqrt(xdiff*xdiff+ydiff*ydiff);
            xmod = -(xdiff*pixelspertick)/totlen;
            ymod = -(ydiff*pixelspertick)/totlen;
            if (weap->getProjectile()->doesRotate()) {
                if( xdiff == 0 ) {
                    if( ydiff < 0 ) {
                        alpha = -3.1415926536f/2.0f;
                    } else {
                        alpha = 3.1415926536f/2.0f;
                    }
                } else {
                    alpha = atan((float)ydiff/(float)xdiff);
                    if( xdiff < 0 ) {
                        alpha = 3.1415926536f+alpha;
                    }
                }
                facing = (40-(Sint8)(alpha*16/3.1415926536f))&0x1f;
                l2o->imagenum = weap->getProjectile()->getImageNum()+facing;

            }
        }
    }
    //check if we are close enough to target to stop modifying
    if( fabsl(xdiff) < fabsl(xmod) ) {
        xmod = 0;
    }
    if( fabsl(ydiff) < fabsl(ymod) ) {
        ymod = 0;
    }
    // if we are so close both
    if( xmod == 0 && ymod == 0 ) {
        //projectile hit..
        if( weap->getWarhead()->getExplosionsound() != NULL ) {
            uspool->getsfxeng()->queueSound(weap->getWarhead()->getExplosionsound());
        }
        new ExplosionAnim(1, uspool, dest, weap->getWarhead()->getEImage(), weap->getWarhead()->getESteps());
        utarget = uspool->getUnitAt(dest, subdest);

        if( utarget != NULL ) {
            if (((UnitType*)utarget->getType())->isInfantry() && inaccurate) {
                for (int sud=0;sud<5;++sud) {
                    utarget = uspool->getUnitAt(dest, sud);
                    if (utarget != NULL) {
                        // each soldier in that cell gets one third of
                        // normal damage
                        utarget->applyDamage((Sint16)((double)weap->getDamage()/3.0),weap);
                    }
                } // targeted soldier gets full normal damage
                utarget = uspool->getUnitAt(dest, subdest);
                if (utarget != NULL) { // soldier might have already been killed
                    utarget->applyDamage((Sint16)(2.0*(double)weap->getDamage()/3.0),weap);
                }
                delete this;
                return;
            } else {
                utarget->applyDamage(weap->getDamage(),weap);
                delete this;
                return;
            }
        }
        starget = uspool->getStructureAt(dest,weap->getWall());
        if( starget != NULL ) {
            starget->applyDamage(weap->getDamage(),weap);
            delete this;
            return;
        }
        //printf("missed target\n");
        delete this;
        return;
    }
    if (!heatseek) {
        // decrease xdiff by xmod and ydiff by ymod
        xdiff += xmod;
        ydiff += ymod;
    }
    // move the actual projectile
    rxoffs += xmod;
    ryoffs += ymod;

    //l2o->xoffset += xmod;
    //l2o->yoffset += ymod;
    oldpos = l2o->cellpos;
    while( rxoffs < 0 ) {
        rxoffs += 24;
        l2o->cellpos--;
    }
    while(rxoffs >= 24) {
        rxoffs -= 24;
        l2o->cellpos++;
    }

    while( ryoffs < 0 ) {
        ryoffs += 24;
        l2o->cellpos -= uspool->getMap()->getWidth();
    }
    while( ryoffs >= 24 ) {
        ryoffs -= 24;
        l2o->cellpos += uspool->getMap()->getWidth();
    }
    l2o->xoffset = (Sint8)rxoffs;
    l2o->yoffset = (Sint8)ryoffs;

    //printf("offs: %lf, %lf\n", rxoffs, ryoffs);
    //printf("offs: %d, %d\n", l2o->xoffset, l2o->yoffset);
    if( oldpos != l2o->cellpos ) {
        uspool->removeL2overlay(l2entry);
        l2entry = uspool->addL2overlay(l2o->cellpos, l2o);
    }
    setDelay(1);
    uspool->scheduleUSEvent(this);
}
