/*****************************************************************************
 * args.cpp - a class that parses command line (cl) arguments and stores
 *            the values parsed-out
 ****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "SDL.h"
#include "freecnc.h"
#include "args.h"
#include "inifile.h"
#include "memtrack.h"

/** Print the help message */
static void PrintUsage() {
    printf("FreeCNC - %s\n\n", VERSION);
    printf("Usage: freecnc [OPTIONS]\n");
    printf("\t-map mapname\tName of mission to load\n");
    printf("\t-w width\tWidth of screen\n");
    printf("\t-h height\tHeight of screen\n");
    printf("\t-bpp bpp\tVideo Depth\n");
    printf("\t-fullscreen\tUse fullscreen mode\n");
    printf("\t-window\tUse windowed mode\n");
    //    printf("\t-ra\tPlay Red Alert\n");
    //    printf("\t-demo\tPlay with the C&C Demo data files\n\n");
    printf("\t-nosound\tPlay without sound\n");
    printf("\t-playvqa vqaname\tPlays a VQA\n\n");
    printf("\t-grab\tGrabs mouse input (locks mouse inside freecnc window)\n");
}

/** function to parse command line arguments
 * @param number of arguments
 * @param pointer to array of arguments
 * @return 0 on success, -1 if user entered invalid parameters
 */
int Args::parse(int argc, char **argv) {
    int i;
    char *tmp;
    INIFile *conquer_ini;

    try {
        conquer_ini = new INIFile("conquer.ini");
    } catch(INInotFound *e) {
        fprintf(stderr, "The inifile \"conquer.ini\" was not found!\n");
        delete e;

        return -1;
    }

    /* Setup defaults */
    /* Some of the "defaults" are in conquer.ini */
    videoflags = SDL_HWSURFACE|SDL_DOUBLEBUF;
#if defined _WIN32 && !defined _DEBUG

    videoflags |= SDL_FULLSCREEN;
#endif

    tmp = conquer_ini->readString("Options", "Map");
    if( tmp == NULL ) {
        fprintf(stderr, "Option \"Map\" missing in inifile\n");
        delete conquer_ini;
        return -1;
    }
    strcpy(map, tmp);
    delete[] tmp;
    Width = conquer_ini->readInt("Video", "Width", 640);
    Height = conquer_ini->readInt("Video", "Height", 480);
    Bpp = conquer_ini->readInt("Video", "Bpp", 16);
    intro = conquer_ini->readInt("Options", "PlayIntro", 1);
    gamenum = conquer_ini->readInt("Options", "Game",GAME_TD);
    nosound = (bool)conquer_ini->readInt("Options", "Nosound",0);
    playvqa = false;
    grabmode = SDL_GRAB_OFF;
    delete conquer_ini;

    for (i = 1; i < argc; i++) {
        if ( strcmp(argv[i], "-nosound") == 0) {
            nosound = true;
        }
        if ( strcmp(argv[i], "-fullscreen") == 0 )
            videoflags |= SDL_FULLSCREEN;
        if ( strcmp(argv[i], "-window") == 0 )
            videoflags &= ~SDL_FULLSCREEN;

        if ( strcmp(argv[i], "-ra") == 0 ) {
            gamenum = GAME_RA;
        }

        if (strcmp(argv[i], "-playvqa") == 0) {
            if (argv[i+1]) {
                playvqa = true;
                strncpy(&vqamovie[0],argv[i+1],15);
                i++;
            }
        }

        if (strcmp(argv[i], "-grab") == 0) {
            grabmode = SDL_GRAB_ON;
        }

        if ( strcmp(argv[i], "-map") == 0 ) {
            if (argv[i+1]) {
                strncpy(map, argv[i+1],8);
                i++;
            }
        }

        if ( strcmp(argv[i], "-w") == 0 ) {
            if (argv[i+1]) {
                Width = atoi(argv[i+1]);
                i++;
            }
        }

        if ( strcmp(argv[i], "-h") == 0 ) {
            if (argv[i+1]) {
                Height = atoi(argv[i+1]);
                i++;
            }
        }

        if ( strcmp(argv[i], "-bpp") == 0 ) {
            if (argv[i+1]) {
                Bpp = atoi(argv[i+1]);
                i++;
            }
        }

        if ( strcmp(argv[i], "-help") == 0 || strcmp(argv[i], "--help") == 0)  {
            /* -help prints the help message and returns -1 to stop execution*/
            PrintUsage();
            return -1;
        }
    }
    return 0;
}
