#include "SDL_endian.h"
#include "vfs.h"
#include "inifile.h"
#include "mixvfs.h"
#include "memtrack.h"

#include <vector>
#include <map>
#include <string.h>

#define MIN(x,y) ((x<y)?x:y)

using namespace std;

VFile::VFile(FILE *file_to_load, Uint32 startpos, Uint32 filesize, bool externalFile) {
    this->size = filesize;
    this->startpos = startpos;
    this->datafile = file_to_load;
    this->externalFile = externalFile;
    pos = 0;
}

VFile::~VFile() {
    if( externalFile ) {
        fclose(datafile);
    }
}

Uint32 VFile::readByte(Uint8 *databuf, Uint32 numBytes) {
    Uint32 numRead;
    numRead = MIN(numBytes, size-pos);
    fseek(datafile, startpos+pos, SEEK_SET);
    numRead = fread(databuf, 1, numRead, datafile);
    pos += numRead;
    return numRead;
}

Uint32 VFile::readWord(Uint16 *databuf, Uint32 numWords) {
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = MIN(numWords, ((size-pos)>>1));
    fseek(datafile, startpos+pos, SEEK_SET);
    numRead = fread(databuf, 2, numRead, datafile);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        SDL_Swap16(databuf[i]);
    }
#endif
    pos += numRead<<1;
    return numRead;
}

Uint32 VFile::readThree(Uint32 *databuf, Uint32 numThrees) {
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = MIN(numThrees, ((size-pos)/3));
    fseek(datafile, startpos+pos, SEEK_SET);
    numRead = fread(databuf, 3, numRead, datafile);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        SDL_Swap32(databuf[i]);
        databuf[i]<<8;
    }
#endif
    pos += numRead*3;
    return numRead;
}

Uint32 VFile::readDWord(Uint32 *databuf, Uint32 numDWords) {
    Uint32 numRead;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    Uint32 i;
#endif

    numRead = MIN(numDWords, ((size-pos)>>2));
    fseek(datafile, startpos+pos, SEEK_SET);
    numRead = fread(databuf, 4, numRead, datafile);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN

    for( i = 0; i < numRead; i++ ) {
        SDL_Swap32(databuf[i]);
    }
#endif
    pos += numRead<<2;
    return numRead;
}

char *VFile::getLine(char *string, Uint32 buflen) {
    Uint32 numRead;
    char *retval;

    numRead = MIN(buflen-1, size-pos);
    if( numRead == 0 ) {
        return NULL;
    }
    fseek(datafile, startpos+pos, SEEK_SET);
    retval = fgets(string, numRead+1, datafile);
    pos += strlen(string);
    return retval;
}


std::vector<char *> paths;
MIXFiles *mixes;

FILE *VFS_getFile(char *fname, Uint32 *startpos, Uint32 *size, bool *external) {
    FILE *fileptr;
    char filename[1024];
    Uint32 i;
    //check if file exists
    for( i = 0; i < paths.size(); i++ ) {
        strcpy(filename, paths[i]);
        strcat(filename, fname);
        fileptr = fopen(filename, "rb");
        if( fileptr != NULL ) {
            fseek(fileptr, 0, SEEK_END);
            (*startpos) = 0;
            (*size) = ftell(fileptr);
            (*external) = true;
            return fileptr;
        }
    }
    //if not look for it in the mixes
    //search the header..
    (*external) = false;
    return mixes->getOffsetAndSize(fname, startpos, size);
}

void VFS_Init() {
    Uint32 i, required, startpos, size;
    bool external;
    FILE *fptr;
    INIFile *filesini;
    char *defpath = new char[6];
    char *game;
    char tempstr[1024];
    strcpy(defpath, "conf/");
    paths.push_back(defpath);
    try {
        filesini = new INIFile("files.ini");
    } catch(char *fname) {
        fprintf(stderr, "Unable to locate \"%s\"\n", fname);
        exit(1);
    }
    i = 1;
    while(defpath != NULL) {
        sprintf(tempstr, "path%d", i++);
        defpath = filesini->readString("general", tempstr);
        if( defpath != NULL ) {
            if( defpath[strlen(defpath)-1] != '/' &&
                    defpath[strlen(defpath)-1] != '\\' ) {
                strcpy(tempstr, defpath);
                strcat(tempstr, "/");
                delete[] defpath;
                defpath = new char[strlen(tempstr)+1];
                strcpy(defpath, tempstr);
            }
            paths.push_back(defpath);
        }
    }

    /* load all the mixes here */
    for( i = 1;;i++ ) {
        mixes = new MIXFiles();
        sprintf(tempstr, "game%d", i);
        game = filesini->readString("general", tempstr);
        if( game == NULL ) {
            break;
        }
        filesini->seekSection(game);
        delete[] game;
        while( game != NULL ) {
            game = filesini->getLine();
            if( game != NULL ) {
                sscanf(game, "%[^=]=%d", tempstr, &required );
                fptr = VFS_getFile(tempstr, &startpos, &size, &external);
                if( fptr == NULL ) {
                    if( required ) {
                        fprintf(stderr, "Missing required file \"%s\"\n", tempstr);
                        delete mixes;
                        delete[] game;
                        break;
                    }
                } else {
                    mixes->loadMix(fptr, startpos, external);
                }
                delete[] game;
            }
        }
        if( game == NULL ) {
            delete filesini;
            return;
        }
    }

    delete filesini;
    fprintf(stderr, "Unable to find mixes for any of the specified games!\n");
    exit(1);
}

void VFS_Destroy() {
    Uint32 i;
    for( i = 0; i < paths.size(); i++ ) {
        delete[] paths[i];
    }
    delete mixes;
}

VFile *VFS_Open(char *fname) {
    FILE *fileptr;
    VFile *virtualfile;
    Uint32 startpos, size;
    bool external;
    fileptr = VFS_getFile(fname, &startpos, &size, &external);
    if( fileptr == NULL ) {
        return NULL;
    }
    virtualfile = new VFile(fileptr, startpos, size, external);
    return virtualfile;
}

void VFS_Close(VFile *file) {
    delete file;
}

