/*****************************************************************************
 * video.cpp - All functions to blit something to screen, update screen, 
 *  	       init screen and such goes in this class
 * 
 ****************************************************************************/
#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif
#include <vector>

#include "graphicsengine.h"
#include "input.h"

#include "memtrack.h"

#define MIN(x,y) x<y?x:y

using namespace std;

/** Constructor, inits the sdl graphics and prepares to render maps
 * and draw vqaframes.
 * @param commandline arguments regarding video.
 */
GraphicsEngine::GraphicsEngine( Args *arg ) {
    width = arg->getWidth();
    height = arg->getHeight();

    /* Set the caption, not nessesary but nice =) */
    SDL_WM_SetCaption("FreeCNC", NULL);
    SDL_WM_GrabInput(arg->getGrabMode());
    /* Init the screen */
    screen = SDL_SetVideoMode(width, height, arg->getBpp(), arg->getVideoFlags());

    if (screen == NULL) {
        fprintf(stderr, "Unable to set %dx%d video: %s\n",
                width, height, SDL_GetError());
        throw new VideoError; /* throw an error for the caller to catch */
    }

    imgcache = new ImageCache();

    messages = new MessagePool();
}



/** Destructor, free the memory used by the graphicsengine. */
GraphicsEngine::~GraphicsEngine() {
    delete imgcache;
    delete messages;
}



/** Setup the vars for the current mission.
 * @param the current map.
 * @param the current sidebar.
 * @param the current cursor.
 */
void GraphicsEngine::setupCurrentGame( CnCMap *map, Sidebar *sidebar, Cursor *cursor ) {

    this->map = map;
    this->sidebar = sidebar;
    this->cursor = cursor;

    /* assume the width of tile 0 is correct */
    //   tilewidth = map->getMapTile(0,0)->w;
    tilewidth = map->getMapTile(0)->w;
    imgcache->setImagePool(map->getImagePool());

    //   SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, SHPBase::palette, 0, 256);
    /* this is a new game so clear the scren */
    clearScreen();
}

/** Render a scene complete with map, sidebar and cursor. */
void GraphicsEngine::renderScene() {
    //   int mx, my;
    int i;
    Sint16 xpos, ypos;

    SDL_Surface *curimg, *bgimage;
    SDL_Rect dest, src, udest;

    ImageCacheEntry *images;

    Uint32 tiberium, smudge, overlay, terrain;
    Uint8 numshps;
    Uint32 *unitorstructshps;
    Sint8 *uxoffsets, *uyoffsets;
    Sint16 txoff, tyoff;
    double ratio;
    Unit* tmp_un;

    vector<Uint16> l2overlays;

    static Uint32 whitepix = SDL_MapRGB(screen->format,0xff, 0xff, 0xff);
    static Uint32 greenpix = SDL_MapRGB(screen->format,0, 0xff, 0);
    static Uint32 yellowpix = SDL_MapRGB(screen->format,0xff, 0xff, 0);
    static Uint32 redpix = SDL_MapRGB(screen->format,0xff, 0, 0);
    static Uint32 blackpix = SDL_MapRGB(screen->format,0, 0, 0);

    Uint16 mapWidth, mapHeight, selection;
    Uint32 curpos, curdpos;

    Sint8 xmax, ymax;
    static SDL_Rect oldmouse = {0, 0, 0, 0};
    /* remove the old mousecursor */
    SDL_FillRect(screen, &oldmouse, blackpix);

    drawSidebar();

    /*draw minimap*/
    if (Input::isMinimapEnabled() ) {
        SDL_Surface *minimap = map->getMiniMap();
        dest.w = MIN(sidebar->getTabLocation()->w, minimap->w);
        dest.h = MIN(dest.w, minimap->h);
        dest.x = maparea.x+maparea.w;
        dest.y = maparea.y;
        src.x = map->getXScroll()*6;
        src.y = map->getYScroll()*6;
        src.w = dest.w;
        src.h = dest.h;
        if( src.x + src.w >= minimap->w ) {
            src.x = minimap->w - src.w;
        }
        if( src.y + src.h >= minimap->h ) {
            src.y = minimap->h - src.h;
        }
        SDL_BlitSurface(minimap, &src, screen, &dest);
    }
    SDL_SetClipRect( screen, &maparea );

    mapWidth = MIN(maparea.w/tilewidth, map->getWidth());
    mapHeight = MIN(maparea.h/tilewidth, map->getHeight());

    dest.y = maparea.y;

    xmax = MIN(map->getWidth()-(map->getXScroll()+mapWidth), 4);
    ymax = MIN(map->getHeight()-(map->getYScroll()+mapHeight), 4);

    curpos = map->getScrollPos();
    if (map->toScroll())
        map->doscroll();
    for( ypos = 0; ypos < mapHeight+ymax+2; ypos++ ) {
        dest.x = maparea.x;
        for( xpos = 0; xpos < mapWidth+xmax; xpos++ ) {
            dest.w = tilewidth;
            dest.h = tilewidth;

            if( xpos < mapWidth && ypos < mapHeight ) {

                bgimage = map->getMapTile(curpos);
                if( bgimage != NULL ) {
                    SDL_BlitSurface(bgimage, NULL, screen, &dest);
                }

                smudge = map->getSmudge(curpos);
                if( smudge != 0 ) {
                    images = imgcache->getImage(smudge);
                    SDL_BlitSurface(images->image, NULL, screen, &dest);
                    SDL_BlitSurface(images->shadow, NULL, screen, &dest);
                }

                tiberium = map->getTiberium(curpos);
                if( tiberium != 0 ) {
                    images = imgcache->getImage(tiberium);
                    SDL_BlitSurface(images->image, NULL, screen, &dest);
                }

                overlay = map->getOverlay(curpos);
                if( overlay != 0 ) {
                    images = imgcache->getImage(overlay);
                    SDL_BlitSurface(images->image, NULL, screen, &dest);
                    SDL_BlitSurface(images->shadow, NULL, screen, &dest);
                }

                if( map->getUASP()->hasL2overlay(curpos) ) {
                    l2overlays.push_back(curpos);
                }
            }
            if( ypos > 1 ) {
                dest.y -= (tilewidth<<1);
                curdpos = curpos-(map->getWidth()<<1);
                terrain = map->getTerrain(curdpos, txoff, tyoff);
                if( terrain != 0 ) {
                    images = imgcache->getImage(terrain);
                    src.x = 0;
                    src.y = 0;
                    src.w = images->image->w;
                    src.h = images->image->h;
                    udest.x = dest.x+txoff;
                    udest.y = dest.y+tyoff;
                    udest.w = images->image->w;
                    udest.h = images->image->h;
                    if( udest.x < maparea.x ) {
                        src.x = maparea.x - udest.x;
                        src.w -= src.x;
                        udest.x = maparea.x;
                        udest.w = src.w;
                    }
                    if( udest.y < maparea.y ) {
                        src.y = maparea.y - udest.y;
                        src.h -= src.y;
                        udest.y = maparea.y;
                        udest.h = src.h;
                    }
                    SDL_BlitSurface(images->image, &src, screen, &udest);
                    SDL_BlitSurface(images->shadow, &src, screen, &udest);
                }

                numshps = map->getUASP()->getUnitOrStructureNum(curdpos, &unitorstructshps,
                          &uxoffsets, &uyoffsets);
                if( numshps > 0 ) {
                    for( i  = 0; i < numshps; i++ ) {
                        images = imgcache->getImage(unitorstructshps[i]);
                        src.x = 0;
                        src.y = 0;
                        src.w = images->image->w;
                        src.h = images->image->h;
                        udest.w = images->image->w;
                        udest.h = images->image->h;
                        udest.x = dest.x+uxoffsets[i];
                        udest.y = dest.y+uyoffsets[i];
                        if( udest.x < maparea.x ) {
                            src.x = maparea.x - udest.x;
                            src.w -= src.x;
                            udest.x = maparea.x;
                            udest.w = src.w;
                        }
                        if( udest.y < maparea.y ) {
                            src.y = maparea.y - udest.y;
                            src.h -= src.y;
                            udest.y = maparea.y;
                            udest.h = src.h;
                        }
                        SDL_BlitSurface(images->image, &src, screen, &udest);
                        SDL_BlitSurface(images->shadow, &src, screen, &udest);
                    }
                    delete[] unitorstructshps;
                    delete[] uxoffsets;
                    delete[] uyoffsets;
                }

                selection = map->getUASP()->getSelected(curdpos);
                if( (selection&0xff) != 0 ) {
                    udest.h = 5;
                    if( selection>>8 == 0xff ) {
                        //udest.w = 12;
                        // safe to assume Unit
                        if( selection &1 ) {
                            tmp_un  = (Unit*)map->getUASP()->getUnitOrStructureAt(curdpos,0);
                            ratio   = tmp_un->getRatio();
                            udest.x = dest.x + 6 + tmp_un->getXoffset();
                            udest.y = dest.y + 5 + tmp_un->getYoffset();
                            udest.w = 12;
                            udest.h = 5;
                            SDL_FillRect(screen, &udest, blackpix);
                            udest.w = (Uint16)(11.0 * ratio);
                            udest.h -= 2;
                            ++udest.x;
                            ++udest.y;
                            SDL_FillRect(screen, &udest, ((ratio<=0.5)?(ratio<=0.25?redpix:yellowpix):greenpix));
                        }
                        if( selection &2 ) {
                            tmp_un  = (Unit*)map->getUASP()->getUnitOrStructureAt(curdpos,1);
                            ratio   = tmp_un->getRatio();
                            udest.x = dest.x + tmp_un->getXoffset();
                            udest.y = dest.y + tmp_un->getYoffset();
                            udest.w = 12;
                            udest.h = 5;
                            SDL_FillRect(screen, &udest, blackpix);
                            udest.w = (Uint16)(11.0 * ratio);
                            udest.h -= 2;
                            ++udest.x;
                            ++udest.y;
                            SDL_FillRect(screen, &udest, ((ratio<=0.5)?(ratio<=0.25?redpix:yellowpix):greenpix));
                        }
                        if( selection &4 ) {
                            tmp_un = (Unit*)map->getUASP()->getUnitOrStructureAt(curdpos,2);
                            ratio   = tmp_un->getRatio();
                            udest.x = dest.x + 12 + tmp_un->getXoffset();
                            udest.y = dest.y + tmp_un->getYoffset();
                            udest.w = 12;
                            udest.h = 5;
                            SDL_FillRect(screen, &udest, blackpix);
                            udest.w = (Uint16)(11.0 * ratio);
                            udest.h -= 2;
                            ++udest.x;
                            ++udest.y;
                            SDL_FillRect(screen, &udest, ((ratio<=0.5)?(ratio<=0.25?redpix:yellowpix):greenpix));
                        }
                        if( selection &8 ) {
                            tmp_un = (Unit*)map->getUASP()->getUnitOrStructureAt(curdpos,3);
                            ratio   = tmp_un->getRatio();
                            udest.x = dest.x + tmp_un->getXoffset();
                            udest.y = dest.y + 10 + tmp_un->getYoffset();
                            udest.w = 12;
                            udest.h = 5;
                            SDL_FillRect(screen, &udest, blackpix);
                            udest.w = (Uint16)(11.0 * ratio);
                            udest.h -= 2;
                            ++udest.x;
                            ++udest.y;
                            SDL_FillRect(screen, &udest, ((ratio<=0.5)?(ratio<=0.25?redpix:yellowpix):greenpix));
                        }
                        if( selection &16 ) {
                            tmp_un = (Unit*)map->getUASP()->getUnitOrStructureAt(curdpos,4);
                            ratio   = tmp_un->getRatio();
                            udest.x = dest.x + 12 + tmp_un->getXoffset();
                            udest.y = dest.y + 10 + tmp_un->getYoffset();
                            udest.w = 12;
                            udest.h = 5;
                            SDL_FillRect(screen, &udest, blackpix);
                            udest.w = (Uint16)(11.0 * ratio);
                            udest.h -= 2;
                            ++udest.x;
                            ++udest.y;
                            SDL_FillRect(screen, &udest, ((ratio<=0.5)?(ratio<=0.25?redpix:yellowpix):greenpix));
                        }

                    } else {
                        ratio = map->getUASP()->getUnitOrStructureAt(curdpos,0)->getRatio();
                        SDL_FillRect(screen, &udest, blackpix);
                        udest.h -= 2;
                        ++udest.x;
                        ++udest.y;
                        udest.w = (Uint16)((double)(udest.w-1) * ratio);
                        SDL_FillRect(screen, &udest, ((ratio<=0.5)?(ratio<=0.25?redpix:yellowpix):greenpix));
                    }

                }

                dest.y += (tilewidth<<1);
            } /* ypos > 1 */
            /* show the different tiles */
            /*       udest.x = maparea.x;
             udest.y = maparea.y+ypos*24;                                           
             udest.h = 1;                                                           
             udest.w = maparea.w;                                                   
             SDL_FillRect(screen, &udest, whitepix);                                
             udest.x = maparea.x+xpos*24;                                           
             udest.y = maparea.y;                                                   
             udest.h = maparea.h;                                                   
             udest.w = 1;                                                           
             SDL_FillRect(screen, &udest, whitepix);
             */

            dest.x += tilewidth;
            curpos++;
        }

        curpos += map->getWidth()-mapWidth-xmax;
        dest.y += tilewidth;
    } /* the for-statements */

    for( i = 0; (Uint32)i < l2overlays.size(); i++ ) {
        curpos = l2overlays[i];
        dest.x = maparea.x+(curpos%map->getWidth()-map->getXScroll())*tilewidth;
        dest.y = maparea.y+(curpos/map->getWidth()-map->getYScroll())*tilewidth;
        numshps = map->getUASP()->getL2overlays(curpos, &unitorstructshps, &uxoffsets, &uyoffsets);
        for( curdpos = 0; curdpos < numshps; curdpos++ ) {
            images = imgcache->getImage(unitorstructshps[curdpos]);
            udest.x = dest.x + uxoffsets[curdpos];
            udest.y = dest.y + uyoffsets[curdpos];
            udest.w = images->image->w;
            udest.h = images->image->h;
            SDL_BlitSurface(images->image, NULL, screen, &udest);
        }
        delete[] unitorstructshps;
        delete[] uxoffsets;
        delete[] uyoffsets;
        //printf("l2 overlay @ %d\n", l2overlays[i]);
    }

    /* draw the selectionbox */
    if( Input::isDrawing() ) {
        dest.x = MIN(Input::getMarkRect().x, Input::getMarkRect().w);
        dest.y = MIN(Input::getMarkRect().y, Input::getMarkRect().h);
        dest.w = abs(Input::getMarkRect().x - Input::getMarkRect().w);
        dest.h = 1;
        SDL_FillRect(screen, &dest, whitepix);
        dest.y += abs(Input::getMarkRect().y - Input::getMarkRect().h);
        SDL_FillRect(screen, &dest, whitepix);
        dest.y -= abs(Input::getMarkRect().y - Input::getMarkRect().h);
        dest.h = abs(Input::getMarkRect().y - Input::getMarkRect().h);
        dest.w = 1;
        SDL_FillRect(screen, &dest, whitepix);
        dest.x += abs(Input::getMarkRect().x - Input::getMarkRect().w);
        SDL_FillRect(screen, &dest, whitepix);
    }

    dest.x = dest.y = 0;
    dest.w = width;
    dest.h = height;
    SDL_SetClipRect( screen, &dest );

    //   SDL_GetMouseState(&mx, &my);

    //   dest.x = mx;
    //   dest.y = my;
    // Draw messages
    curimg = messages->getMessages();
    if( curimg != NULL ) {
        dest.x = maparea.x;
        dest.y = maparea.y;
        dest.w = curimg->w;
        dest.h = curimg->h;

        SDL_BlitSurface(curimg, NULL, screen, &dest);
    }

    // Draw the mouse
    dest.x = cursor->getX();
    dest.y = cursor->getY();
    if( dest.x < 0 )
        dest.x = 0;
    if( dest.y < 0 )
        dest.y = 0;

    curimg = cursor->getCursor();
    dest.w = curimg->w;
    dest.h = curimg->h;

    //oldmouse = dest;
    SDL_BlitSurface(curimg, NULL, screen, &dest );


    SDL_Flip(screen);
#ifdef _WIN32

    SDL_FillRect(screen, &oldmouse, blackpix);
#endif

    oldmouse = dest;
}



/** Draw the sidebar and tabs. */
void GraphicsEngine::drawSidebar() {
    SDL_Rect dest;
    SDL_Rect *tabpos;

    static Uint32 firstframe = SDL_GetTicks();
    static Uint32 frames = 0;
    //    static Uint32 blackpix = SDL_MapRGB(screen->format, 0, 0, 0);
    Uint32 tick;
    Uint16 framerate;
    char mtext[128];

    /* clear parts of the screen not overwriten by the sidebar or map
     remove this code */
    /*    dest.x = 0;
        dest.y = 0;
        dest.w = width;
        dest.h = maparea.y;
        SDL_FillRect(screen, &dest, blackpix);
        dest.y = maparea.y+maparea.h;
        dest.h = height-dest.y;
        SDL_FillRect(screen, &dest, blackpix);
        dest.y = maparea.y;
        dest.w = maparea.x;
        dest.h = maparea.h;
        SDL_FillRect(screen, &dest, blackpix);
        dest.x = maparea.x+maparea.w;
        dest.w = width - dest.x
        SDL_FillRect(screen, &dest, blackpix);
    */
    tabpos = sidebar->getTabLocation();


    if( sidebar->getVisible() ) {
        if( sidebar->getChanged() ) {
            clearScreen();

            maparea.h = ((height-tabpos->h)/tilewidth)*tilewidth;
            maparea.y = ((height-tabpos->h-maparea.h)>>1)+tabpos->h;
            maparea.w = ((width-tabpos->w)/tilewidth)*tilewidth;
            maparea.x = (width-tabpos->w-maparea.w)>>1;
            map->setMaxScroll(maparea.w/tilewidth, maparea.h/tilewidth);
        }
        dest.x = width - tabpos->w;
        dest.w = tabpos->w;
        dest.y = maparea.y;
        dest.h = maparea.h;

        //      SDL_FillRect(screen, &dest, SDL_MapRGB(screen->format, 0xa0, 0xa0, 0xa0));
        SDL_BlitSurface( sidebar->getSidebarImage(dest), NULL, screen, &dest);
    } else if( sidebar->getChanged() ) {
        clearScreen();

        maparea.h = ((height-tabpos->h)/tilewidth)*tilewidth;
        maparea.y = ((height-tabpos->h-maparea.h)>>1)+tabpos->h;
        maparea.w = (width/tilewidth)*tilewidth;
        maparea.x = (width-maparea.w)>>1;
        map->setMaxScroll(maparea.w/tilewidth, maparea.h/tilewidth);
    }

    tabpos->y = maparea.y-tabpos->h;
    tabpos->x = 0;

    SDL_BlitSurface( sidebar->getTabImage(), NULL, screen, tabpos);

    dest.y = tabpos->y+tabpos->h-sidebar->getFont()->getHeight();
    dest.x = ((tabpos->w-sidebar->getFont()->calcStringWidth("Options"))>>1)+tabpos->x;
    sidebar->getFont()->drawString("Options", screen, dest.x, dest.y);

    tabpos->x = width - tabpos->w;

    SDL_BlitSurface( sidebar->getTabImage(), NULL, screen, tabpos);
    sprintf(mtext, "%d", map->getPlayerPool()->getHPlayer()->getMoney());
    dest.x = ((tabpos->w-sidebar->getFont()->calcStringWidth(mtext))>>1)+tabpos->x;
    sidebar->getFont()->drawString(mtext, screen, dest.x, dest.y);


    tick = SDL_GetTicks();
    frames++;
    framerate = (frames*1000)/(tick - firstframe+1);
    sprintf(mtext, "%d fps", framerate );
    SDL_Rect clearCoor;
    clearCoor.w = 100;
    clearCoor.h = 20;
    clearCoor.x = width >>1;
    clearCoor.y = dest.y;
    SDL_FillRect(screen, &clearCoor, SDL_MapRGB(screen->format, 0, 0, 0));
    sidebar->getFont()->drawString(mtext, screen, width>>1, dest.y);
    if( frames == 1000 ) {
        frames = 1;
        firstframe = tick;
    }
}



/** draw a vqafarame to the screen.
 * @param the vqaframe.
 * @param where to draw the frame.
 */
void GraphicsEngine::drawVQAFrame(SDL_Surface *frame, SDL_Rect *dest) {
    SDL_BlitSurface(frame, NULL, screen, dest );
    SDL_Flip(screen);
}



/** Clear the screen, i.e. paint it black. */
void GraphicsEngine::clearScreen() {
    static Uint32 blackpix = SDL_MapRGB(screen->format, 0, 0, 0);
    SDL_Rect dest;

    dest.x = 0;
    dest.y = 0;
    dest.w = width;
    dest.h = height;

    SDL_FillRect(screen, &dest, blackpix);

    /* directx needs to clear both front and back buffer */

#ifdef _WIN32

    SDL_Flip(screen);
    SDL_FillRect(screen, &dest, blackpix);
#endif

}

